/* vec.c - Linienzge aus Bitmap interpretieren - Copyright 2004 Tobias Becker */

/*
#define DEBUG 1
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* fuer memset() */

typedef struct {
	unsigned short header;
	unsigned short xanz;
	unsigned short yanz;
	unsigned short xoffset;
	unsigned char  *daten;
} bmp_t;

static bmp_t global_bmp;

void scan (unsigned char *dat, bmp_t *bmp, FILE *fd, int allpix);

int such_richtung (int diffs[], unsigned char *von, unsigned char *zu);
char *xystr(unsigned char *pos);
int xy_aus_addr (unsigned short *x, unsigned short *y, unsigned char *p, unsigned short l, unsigned char *anf);
void enden(unsigned char **anfang, unsigned char **ende, unsigned char *pdat, unsigned short l);
unsigned char *zum_anfang(unsigned char *start, unsigned short l);
int nachbar2 (unsigned char *pixel, unsigned short l, unsigned char **nachbar);
int nachbarn (unsigned char *pixel, unsigned short l);

unsigned char *laden(FILE *fd, bmp_t *bmp);
int bmpinfo(FILE *fd, bmp_t *bmp);

int main (int argc, char *argv[]) {
	unsigned char *dat;
	FILE *in, *out;
	bmp_t bmp;
	int ret, allpix=0;

#ifdef DEBUG
	setvbuf(stdout, NULL, _IONBF, 0);
#endif
	
	if (argc < 3) {
		printf("Aufruf: vec in.bmp out.txt [.]   (. nur Pixel)\n");
		return 1;
	}
	/* Im Binaermodus oeffnen, sonst Farbe 13 == \n ... */
	if ((in = fopen(argv[1],"rb")) == NULL) {
		printf("Kann Quelldatei %s nicht oeffnen!\n", argv[1]);
		return 2;
	}
	if ((out = fopen(argv[2],"w")) == NULL) {
		printf("Kann Zieldatei %s nicht oeffnen!\n", argv[2]);
		return 3;
	}

	/* ggf. alles in Pixeln schreiben. */
	if (argv[3] && *argv[3] == '.')
		allpix = 1;

	/* Bitmap-Dimensionen erkennen */
	if ((ret=bmpinfo(in, &bmp))!=0) {
		printf("Bitmap-Datei unverstaendlich (Code %i)!\n", ret);
		return 4;
	}
	printf("Bitmap: Header %hu, Breite %hu (%hu), Hoehe %hu\n", bmp.header, bmp.xanz, bmp.xoffset, bmp.yanz);

	dat = laden(in, &bmp);
	if (dat == NULL) {
		printf("Fehler beim Laden der Bitmap-Datei!");
		return 5;
	}
	fclose(in);
	/* Zeiger auf Pixel in bmp-Struktur speichern und diese dateiweit verfgbar machen */
	bmp.daten = dat;
	global_bmp = bmp;
	printf("Bitmap geladen.\n");

#ifdef DEBUG
	{
		unsigned char *p = dat;
		while (*p==0)
			++p;
		printf("main: Erster Pixel, Farbe %i gesetzt bei (%s)\n", (int) *p, xystr(p));
	}
#endif
	
	/* Die Datei analysieren und Ausgabe erstellen. */
	scan(dat, &bmp, out, allpix);

	free(dat);
	fclose(out);
	return 0;	
}


void scan (unsigned char *dat, bmp_t *bmp, FILE *fd, int allpix) {
	/* Pixel-Koordinaten */
	unsigned short ypos, xpos;
	/* durchlaufender Zeiger */
	unsigned char *pdat;
	long n;
	int m;
	/* Differenzen in dat fr Pixel in Richtung 789  4 6  123 (Ziffernblock) */
	int diff[10] = {0};
	diff[7] = -bmp->xoffset - 1; /* links oben */
	diff[8] = -bmp->xoffset;     /* oben */
	diff[9] = -bmp->xoffset + 1; /* rechts oben */
	diff[4] = -1;                /* links */
	diff[6] =  1;                /* rechts */
	diff[1] =  bmp->xoffset - 1; /* links unten */
	diff[2] =  bmp->xoffset;     /* unten */
	diff[3] =  bmp->xoffset + 1; /* rechts unten */



	/*
	 * ausserste Pixel zur Sicherheit loeschen, damit bei Pixelvefolgung
	 * nicht auf ausserhalb des Bildes liegende Pixel zugegriffen wird,
	 * auch fr Einzelpixel wichtig, da ansonsten Nachbarn auf dem Rand
	 * liegen.
	 */
	/* oberste Zeile */
	memset(dat, 0, bmp->xoffset);
	/* unterste Zeile */
	memset(dat + bmp->xoffset * (bmp->yanz - 1), 0, bmp->xoffset);
	for (ypos = 1; ypos < bmp->yanz-1; ypos++) {
		*(dat + ypos * bmp->xoffset) = 0; /* links */
		*(dat + ypos * bmp->xoffset + bmp->xanz - 1) = 0; /* rechts */
	}

#ifdef DEBUG
	{
		unsigned char *p = dat;
		while (*p==0)
			++p;
		printf("scan: Nach Randlschung erster Pixel, Farbe %i gesetzt bei (%s)\n", (int) *p, xystr(p));
	}
#endif


	/* 
	 * ueber alle Pixel (ausser am Rand) gesetzte Pixel suchen
	 * 1. Durchlauf Pixel mit einem oder mehr als zwei Nachbarn,
	 * somit bleiben fr den zweiten Durchlauf nur noch Linien uebrig
	 * 
	 * Das ganze fters machen, da nach ersten Durchlauf I zumindest Einzelpixel
	 * brig bleiben knnen, mit denen Durchlauf II nicht klar kommt.
	 */
	m = 0;
	do {
		n = 0;
		m++;
		pdat = dat + bmp->xoffset;
		for (ypos = 1; ypos < bmp->yanz-1; ypos++) {
			/* ersten Pixel der Zeile berspringen */
			pdat++;
			for (xpos = 1; xpos < bmp->xanz-1; xpos++) {
				if (*pdat) {
					/* Wir haben "etwas" gefunden. */
					/* Palettenindex aktueller Pixel */
					unsigned char farbe = *pdat;
					int na = (allpix) ? 0 : nachbarn(pdat, bmp->xoffset);
#if 0 && defined(DEBUG)
					printf("scan: Fund (%s) [%2hu,%2hu], Farbe %i, Nachbarn %i\n",
					       xystr(pdat), xpos, ypos, (int) *pdat, na);
#endif
					if (na == 0) {
						/* einzelner Pixel (Bahnhof) */
						fprintf(fd, "P %u %hu,%hu\n", (int) farbe, xpos, ypos);
						*pdat = 0;
						n++;
					}
					if (na > 2) {
						/* mehr als 2 Nachbarn --> keine Linie */
						fprintf(fd, "M %u %hu,%hu\n", (int) farbe, xpos, ypos);
						*pdat = 0;
						n++;
					}
				}
				pdat++;
			}
			/* letzten Pixel der Zeile berspringen */
			pdat++;
			/* ggf. Padding hinzuaddieren */
			pdat += bmp->xoffset - bmp->xanz;
		}
	
		fflush(fd);
		printf("%i. Durchlauf I fuer Punkte (P/M) abgeschlossen (%li Treffer).\n",
		       m, n);
	}
	while (!allpix && n!=0);
	
	/* 2. Durchlauf II fuer Linien (pdat zuruecksetzen!) */
	pdat = dat + bmp->xoffset;
	for (ypos = 1; ypos < bmp->yanz-1; ypos++) {
		/* ersten Pixel der Zeile berspringen */
		pdat++;
		for (xpos = 1; xpos < bmp->xanz-1; xpos++) {
			if (*pdat) {
				/*
				 * Wir haben eine Linie (hoffentlich) gefunden.
				 * - Anfang und Ende bestimmen
				 * - vom Anfang kantenweise vorgehen
				 */
				 /* Zeiger auf Anfang und Ende der Linie, vorigen Pixel auf Linie und jetzigen Pixel der Linie */
				unsigned char *anfang, *ende, *lpos, *pos, *a[2], farbe;
				/* Pixelkoordinaten */
				unsigned short x, y;
				int richtung = 0, m;
				
				/* Im Kreisfall wird erste Fundstelle gelscht, dann fehlt Farbe */
				farbe = *pdat;
				
#if DEBUG
				printf("scan: Linie, Farbe %i gefunden bei #%lX (%s) [%hu,%hu]\n",
				       (int) *pdat, (unsigned long) (pdat - dat), xystr(pdat), xpos, ypos);			
#endif
				
				enden(&anfang, &ende, pdat, bmp->xoffset);

#if DEBUG
				printf("scan: Linien-Anfang bei #%lX (%s)\n", (unsigned long) (anfang - dat), xystr(anfang));			
				printf("scan: Linien-Ende   bei #%lX (%s)\n", (unsigned long) (ende - dat), xystr(ende));			
#endif

		
				pos = anfang;
				xy_aus_addr(&x, &y, anfang, bmp->xoffset, dat);
				fprintf(fd, "L %u %hu,%hu", (int) farbe, x, y);
				
				/* kommt von alleine "in den Tritt" */
				while (pos != ende) {

					if ( richtung != 0 && *pos != *(pos + diff[richtung]) ) {
						/* Kantenende */
						xy_aus_addr(&x, &y, pos, bmp->xoffset, dat);
						fprintf(fd, "-%i-%hu,%hu", richtung, x, y);
						richtung = 0;
					}
					
					if (richtung == 0) {
						/* Neue Richtung suchen */
						m=nachbar2(pos, bmp->xoffset, a);
						if (m != 1) {
							fprintf(stderr, "scan/Linenverfolgung: %i Nachbarn an Pos. %lu (%s), Farbe %i!\n",
							                 m, (unsigned long) (pos-dat), xystr(pos), (int) *pos);
							return;
						}
						richtung = such_richtung(diff, pos, a[0]);
						if (richtung == 0) {
							fprintf(stderr, "scan: Richtung von %lu zu %lu unbekannt!\n", (unsigned long) (pos-dat), (unsigned long) (a[0]-dat));
							return;
						}
					}
					
					/* einen Pixel weitergehen, vorigen loeschen */
					*pos = 0;
					pos += diff[richtung];
					
				}
				/* Polygonende */
				xy_aus_addr(&x, &y, pos, bmp->xoffset, dat);
				fprintf(fd, "-%i-%hu,%hu\n", richtung, x, y);
				*pos = 0;

			} /* /Pixel nicht leer */
			pdat++;
		} /* /Schleife x */
		/* letzten Pixel der Zeile berspringen */
		pdat++;
		/* ggf. Padding hinzuaddieren */
		pdat += bmp->xoffset - bmp->xanz;
	} /* Schleife y */

} /* /scan() */

char *xystr(unsigned char *pos) {
	static char string[8][64];
	static int n=7;
	unsigned short x,y;

	n = ++n%8;
	xy_aus_addr(&x, &y, pos, global_bmp.xoffset, global_bmp.daten);
	snprintf(string[n], 64, "%2hu,%2hu", x, y);
	return string[n];
}



/* aus Differenz die Richtung bestimmen */
int such_richtung (int diffs[], unsigned char *von, unsigned char *zu) {
	int diff = zu - von;
	int i;
	for (i=9; i>0; i--) {
		if (diff == diffs[i])
			break;
	}
	return i;
}
	
void enden(unsigned char **anfang, unsigned char **ende, unsigned char *mitte, unsigned short l) {
	unsigned char mittenwert = *mitte, *a[2];
	int n;
	n = nachbar2(mitte, l, a);
#if 1 && defined(DEBUG)
	printf("enden: %i Nachbarn fuer (%s)", n, xystr(mitte));
	if (n>0)
		printf(" -- (%s)", xystr(a[0]));
	if (n>1)
		printf(", (%s)", xystr(a[1]));
	if (n>2)
		printf(", ...");
	printf("\n");
#endif
	*mitte = 0;
	*anfang = zum_anfang(a[0], l);
	if (n==2)
		*ende = zum_anfang(a[1], l);
	else
		*ende = mitte;
	*mitte = mittenwert;

	/* Kreise erkennen --> Mitten-Pixel auslassen */
	if (2==nachbar2(mitte, l, a) &&
	    ( a[0]==*anfang && a[1]==*ende || a[0]==*ende && a[1]==*anfang ) ) {
#if defined(DEBUG)
		printf("enden: Kreis!\n");
#endif
		*mitte = 0;
	}

#if defined(DEBUG)
	printf("enden: Ergebnis: mitte=(%s), anfang=(%s), ende=(%s), n[mitte]=%i\n",
	       xystr(mitte), xystr(*anfang), xystr(*ende), n);
#endif
	
	
}


/* Linie in eine Richtung bis zum Anfang verfolgen */
unsigned char *zum_anfang(unsigned char *start, unsigned short l) {
	unsigned char *pos = start, *lpos=start;
	unsigned char *a[2];
	/* Nachbarn von Anfangspunkt finden, mit erstem und einzigen gefundenen weitermachen. */
	if (0 == nachbar2(pos, l, a))
		/* bei 2-Pixel-Linie wird 1. Pixel in enden() temporaer geloescht, 2. hat dann hier keine Nachbarn mehr */
		return start;
	pos = a[0];
	/* nun immer weiter in diese "Richtung" */
	while (nachbar2(pos, l, a) == 2) {
		if (a[0] == lpos) {
			lpos = pos;
			pos = a[1];
		}
		else if (a[1] == lpos){
			lpos = pos;
			pos = a[0];
		}
		else {
			fprintf(stderr, "zum_anfang: nachbar2 liefert nicht lpos!\n");
			return start;
		}
	}

	/* pos ist Linienende, wenn nicht haben wir ein Problem */
	return pos;
}

/* Koordinaten aus einem Zeiger bestimmen */
int xy_aus_addr (unsigned short *x, unsigned short *y, unsigned char *p, unsigned short l, unsigned char *anf) {
	unsigned long diff = p - anf;
	*x = diff % l;
	*y = diff / l;
	return 0;
}


/* Abwandlung von nachbarn: Adressen von bis zu zwei Nachbarpixeln liefern */
int nachbar2 (unsigned char *pixel, unsigned short l, unsigned char **nachbar) {
	unsigned char *p;
	int n=0;

#if 1
#ifdef DEBUG
	printf("nachbar2: (%s)=5, Nachbarn dazu (l=%li): ", xystr(pixel), (long) l);
#endif
#endif

	/* links oben beginnen */
	p = pixel - l - 1;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 7");
#endif
	}
	p++;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 8");
#endif
	}
	p++;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 9");
#endif
	}
	/* links in der mittleren Zeile */
	p = pixel - 1;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 4");
#endif
	}
	p += 2;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 6");
#endif
	}
	/* untere Zeile links */
	p = pixel + l - 1;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 1");
#endif
	}
	p++;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 2");
#endif
	}
	p++;
	if (*p == *pixel) {
		if (n < 2)
			*(nachbar + n) = p;
		n++;
#ifdef DEBUG
		printf(" 3");
#endif
	}

#ifdef DEBUG
	printf(", insgesamt %i Nachbarn\n", n);
#endif

	return n;
}


/* Zahl der benachbarten gleichfarbigen Pixel bestimmen */
int nachbarn (unsigned char *pixel, unsigned short l) {
	unsigned char *p;
	int n = 0;

	/* links oben beginnen */
	p = pixel - l - 1;
	if (*p == *pixel)
		n++;
	p++;
	if (*p == *pixel)
		n++;
	p++;
	if (*p == *pixel)
		n++;
	/* links in der mittleren Zeile */
	p = pixel - 1;
	if (*p == *pixel)
		n++;
	p += 2;
	if (*p == *pixel)
		n++;
	/* untere Zeile links */
	p = pixel + l - 1;
	if (*p == *pixel)
		n++;
	p++;
	if (*p == *pixel)
		n++;
	p++;
	if (*p == *pixel)
		n++;
	return n;
}



unsigned char *laden(FILE *fd, bmp_t *bmp) {
	unsigned char *pixel;
	unsigned long ipos, opos;

	pixel = malloc((size_t) bmp->xoffset * bmp->yanz);
	if (pixel==NULL)
		return pixel;
		
	rewind(fd);
	
	/* oberste Zeile steht am Ende der Datei ..*/
	opos = 0;
	ipos = bmp->header;
	ipos += (unsigned long)(bmp->yanz - 1) * bmp->xoffset;
	while (ipos >= (unsigned long) bmp->header) {
		if (fseek(fd, ipos, 0) != 0 || fread(pixel+opos, bmp->xoffset, 1, fd) != 1)
			break;
		ipos -= bmp->xoffset;
		opos += bmp->xoffset;
	}
	if (opos != (unsigned long) bmp->xoffset * bmp->yanz) {
		fprintf(stderr, "laden: Einlesevorgang unvollstaendig, ipos=%lX opos=%lX\n", ipos, opos);
		free (pixel);
	}
	
	return pixel;	
}


int bmpinfo(FILE *fd, bmp_t *bmp) {
	unsigned long fsize;
	unsigned char buf[16];

	rewind(fd);

	if (fread(buf, 2, 1, fd) != 1 || buf[0] != 'B' || buf[1] != 'M')
		return 1;
	
	if (fread(buf, 4, 1, fd) != 1)
		return 2;
	/* Dateigroesse */
	fsize = *((long *) buf);
	
	/* reserved berspringen */
	fseek(fd, 4, 1);
	
	if (fread(buf, 4, 1, fd) != 1)
		return 3;
	/* Headerlaenge */
	bmp->header = (unsigned short) *((long *) buf);
	
	/* Header Size berspringen */
	fseek(fd, 4, 1);

	if (fread(buf, 4, 1, fd) != 1)
		return 4;
	/* Breite (Pixel) */
	bmp->xanz = (unsigned short) *((long *) buf);
	/* In RGB encoding (no compression), when using 8 bits per pixel, lines
	 * must start on a long-word boundary (i.e., low two bits zero).
     */
    bmp->xoffset = 4*((bmp->xanz+3)/4);

	if (fread(buf, 4, 1, fd) != 1)
		return 5;
	/* Hoehe (Pixel) */
	bmp->yanz = (unsigned short) *((long *) buf);

	if ( ((unsigned long)bmp->xoffset) * bmp->yanz + bmp->header != fsize )
		return 6;
	
	return 0;
}

